// Author: Steven Wolfman
// Assorted additions by Alan Hu...
// This work is licensed under a Creative Commons Attribution 3.0 Unported License.
// 
// CPSC 221 students: log into a random linXX.ugrad.cs.ubc.ca machine
// (like lin22.ugrad.cs.ubc.ca.. but if you pick that one, so will everyone else!).
// Then download the Makefile that came with this.
//
// You should be able to compile by simply typing "make".

#include <ctime>     // seed rand (srand) with time
#include <cstdlib>   // srand, rand
#include <cassert>   // assert
#include <vector>    // vector
#include <chrono>    // duration
#include <algorithm> // max (and I recommend learning for_each!)
#include <thread>    // thread!!
#include <iostream>  // cout
#include <omp.h>     // OpenMP functions (get pragmas by compiling w/-fopenmp)
#include "Timer.hpp" // a chrono-based high-precesion timer

// Number of repetitions to perform to get useful timing.
#define REPS 100



void usage(char * name) {
  std::cout << name << " <n>" << std::endl;
  std::cout << "\tCreates an array of <n> random numbers 0-<n>/10" << std::endl
	    << "\tand counts how many are 1s, reporting the time." 
	    << std::endl << std::endl;
  std::cout << "(To make timing clearer, repeats " << REPS 
	    << " times and reports total time taken.)" << std::endl;
}


long long count_matches(long long array[], long long len, long long target) {
  long long matches = 0;
  for (long long i = 0; i < len; i++) {
    if (array[i] == target)
      matches++;
  }
  return matches;
}

// Just calls the sequential counter with appropriate arguments, 
// placing the result into shared memory.
void cmp_helper(long long * result, long long array[],
		long long lo, long long hi, long long target) {
  *result = count_matches(array + lo, hi - lo, target);
}


long long cm_parallel(long long array[], 
		      long long len, 
		      long long target) {
  int divisions = 4;  // How many ways to divide up the array

  // Warning: there's deep magic happening that we can initialize
  // the threads here and assign into them below and have it all work.
  // Ask me if you want to know more, but BEWARE.  (Involves passing by
  // "rvalue reference" using the && type modifier.)
  std::thread workers[divisions]; 
  long long results[divisions];
  for (int d = 0; d < divisions; d++) {
    // Notice that we share memory here by passing a pointer to
    // results[d].  It's tempting to just make it a reference
    // parameter.  Give it a try and see what happens.  (The
    // std::thread constructor takes a strange kind of parameter
    // called an rvalue reference, which is why you get surprising
    // results.)
    workers[d] = 
      std::thread(&cmp_helper, &results[d], array, 
		  (d*len)/divisions, ((d+1)*len)/divisions, target);
  }

  long long matches = 0;
  for (int d = 0; d < divisions; d++) {
    // Buggy version!
    // workers[d].join();
    matches += results[d];
  }

  return matches;
}

// --- Zombie Apocalypse OMP Version Here
// This is the example from class, using divide-and-conquer to
// create lots of threads quickly.
long long zombie_helper(long long array[], 
		      long long len, 
		      long long target) {
    long long left;
    long long right;

    // In real life, you should define a sequential cut-off constant...
    if (len < 1000) return count_matches(array,len,target);

    #pragma omp task untied shared(left)
    left = zombie_helper(array, len/2, target);
    right = zombie_helper(array+len/2, len-(len/2), target);
    #pragma omp taskwait

    return left+right;
}


long long cm_zombie(long long array[], 
		      long long len, 
		      long long target) {
	long long result;

	#pragma omp parallel
	#pragma omp single
	result = zombie_helper(array, len, target);

	return result;
}
// --- End Zombie Apocalypse

// --- Zombie Apocalypse with C++11 Threads
// I didn't do this in lecture, but it's fun to try doing our
// Zombie apocalypse version with C++11 threads.  It's easy to do,
// but because C++11 threads are much more heavyweight than OMP tasks,
// you won't see speedup unless you make the sequential cut-off a lot bigger.
// Try making it a lot smaller and see what happens.

void zombie_helper_cppthreads(long long *result,
		      long long array[], 
		      long long len, 
		      long long target) {
    long long left;
    long long right;

    // In real life, you should define a sequential cut-off constant...
    if (len < 100000) {
    	*result = count_matches(array,len,target);
	return;
    }

    std::thread foo = std::thread(&zombie_helper_cppthreads,
    					&left, array, len/2, target);
    zombie_helper_cppthreads(&right, array+len/2, len-(len/2), target);
    foo.join();

    *result = left+right;
}


long long cm_zombie_cppthreads(long long array[], 
		      long long len, 
		      long long target) {
	long long result;

	zombie_helper_cppthreads(&result, array, len, target);

	return result;
}

// --- End Zombie Apocalypse with C++11 Threads


// Run a count_match algorithm and report the timing/results to standard output.
// Additionally, return the calculated value.
long long runAndReport(std::function<long long(long long*, long long, long long)> f,
		       long long * array, long long len, long long target,
		       const char * type) {
  long long matches;
  Timer timer;
  for (int i = 0; i < REPS; i++) {
    long long result = f(array, len, target);
    assert(i == 0 || result == matches);
    matches = result;
  }
  auto time = timer.timeSoFar();
  std::chrono::duration<double> timeS = time;
  std::cout << type << "\t" << matches << "\t"
	    << timeS.count() << std::endl;
  std::cout.flush();  

  return matches;
}

int main(int argc, char * argv[]) {
  // Initialize the random number generator.
  auto seed = time(NULL);
  srand((unsigned int)seed);

  // Check and process the argument.
  if (argc != 2) {
    usage(argv[0]);
    exit(-1);
  }
  long long n = atoi(argv[1]);

  // Generate the array.  Allocated dynamically so we can try BIG arrays.
  // The stack is blown too easily! 
  long long * array = new long long[n];
  for (long long i = 0; i < n; i++) {
    array[i] = rand() % (n/10) + 1;
  }

  std::cout << "Type\t\t\t\tCount\tTime (s)" << std::endl;
  std::cout.flush();

  long long seqlMatches = runAndReport(&count_matches, array, n, 1,
				       "Sequential              ");
  long long paraMatches = runAndReport(&cm_parallel, array, n, 1,
				       "Constant C++11 threads  ");
  /*
  long long zombieMatches = runAndReport(&cm_zombie, array, n, 1,
				       "Zombie                  ");
  long long zombC11Matches = runAndReport(&cm_zombie_cppthreads, array, n, 1,
				       "Zombie C++11            ");
  */

  assert(paraMatches == seqlMatches);

  // So, how many processors are there?  Try these out..
  std::cout << std::endl;
  std::cout << "C++11  reports " << std::thread::hardware_concurrency() 
	    << " processors." << std::endl;
  std::cout << "OpenMP reports " << omp_get_num_procs() 
	    << " processors." << std::endl;

  delete [] array;
  array = NULL;
  return 0;
}
